"""Declare data objects.

"""
from libcpp.string cimport string
from libcpp.vector cimport vector

cimport numpy as np


cdef extern from "include/dataobjs.hpp":
    # --------------------------------------------------------------------
    # Binning schemes
    # --------------------------------------------------------------------

    cdef cppclass CppBinning "trv::Binning":
        string space
        string scheme
        double bin_min
        double bin_max
        int num_bins
        vector[double] bin_edges
        vector[double] bin_centres
        vector[double] bin_widths

        CppBinning(string space, string scheme) except +

        void set_bins(double bin_min, double bin_max, int nbins)
        void set_bins(double boxsize_max, int ngrid_min)


    # --------------------------------------------------------------------
    # Line of sight
    # --------------------------------------------------------------------

    struct LineOfSight "trv::LineOfSight":
        double pos[3]


    # --------------------------------------------------------------------
    # Clustering statistics
    # --------------------------------------------------------------------

    # -- Two-point statistics --------------------------------------------

    struct PowspecMeasurements "trv::PowspecMeasurements":
        vector[double] kbin
        vector[double] keff
        vector[int] nmodes
        vector[np.complex128_t] pk_raw
        vector[np.complex128_t] pk_shot

    struct TwoPCFMeasurements "trv::TwoPCFMeasurements":
        vector[double] rbin
        vector[double] reff
        vector[int] npairs;
        vector[np.complex128_t] xi

    struct TwoPCFWindowMeasurements "trv::TwoPCFWindowMeasurements":
        vector[double] rbin
        vector[double] reff
        vector[int] npairs;
        vector[np.complex128_t] xi

    # -- Three-point statistics ------------------------------------------

    struct BispecMeasurements "trv::BispecMeasurements":
        vector[double] k1_bin
        vector[double] k2_bin
        vector[double] k1_eff
        vector[double] k2_eff
        vector[int] nmodes
        vector[np.complex128_t] bk_raw
        vector[np.complex128_t] bk_shot

    struct ThreePCFMeasurements "trv::ThreePCFMeasurements":
        vector[double] r1_bin
        vector[double] r2_bin
        vector[double] r1_eff
        vector[double] r2_eff
        vector[int] npairs
        vector[np.complex128_t] zeta_raw
        vector[np.complex128_t] zeta_shot

    struct ThreePCFWindowMeasurements "trv::ThreePCFWindowMeasurements":
        vector[double] r1_bin
        vector[double] r2_bin
        vector[double] r1_eff
        vector[double] r2_eff
        vector[int] npairs
        vector[np.complex128_t] zeta_raw
        vector[np.complex128_t] zeta_shot


cdef class Binning:
    cdef CppBinning* thisptr
    cdef public str space
    cdef public str scheme
    cdef public double bin_min
    cdef public double bin_max
    cdef public int num_bins
    cdef public vector[double] bin_edges
    cdef public vector[double] bin_centres
    cdef public vector[double] bin_widths



"""Declare :cpp:class:`trv::ParticleCatalogue` and its
data-loading method.

"""
from libcpp.vector cimport vector


cdef extern from "include/particles.hpp":
    cdef cppclass CppParticleCatalogue "trv::ParticleCatalogue":
        CppParticleCatalogue(int verbose)

        # int initialise_particles(const int num)
        # int finalise_particles()

        int load_particle_data(
            vector[double] x, vector[double] y, vector[double] z,
            vector[double] nz, vector[double] ws, vector[double] wc
        ) except +


cdef class _ParticleCatalogue:
    cdef CppParticleCatalogue* thisptr



"""
Parameter Set (:mod:`~triumvirate.parameters`)
==========================================================================

Configure program parameter set.

.. autosummary::
    NonParameterError
    InvalidParameterError
    ParameterSet
    fetch_paramset_template

"""
from pathlib import Path
from pprint import pformat, pprint

import numpy as np
import yaml

from .parameters cimport CppParameterSet


# NOTE: Nested entries must not have a non-None default value.
_TMPL_PARAM_DICT = {
    'directories': {
        'catalogues': None,
        'measurements': None,
    },
    'files': {
        'data_catalogue': None,
        'rand_catalogue': None,
    },
    'tags': {
        'output': None,
    },
    'boxsize': {'x': None, 'y': None, 'z': None},
    'ngrid': {'x': None, 'y': None, 'z': None},
    'alignment': 'centre',
    'padscale': 'box',
    'padfactor': None,
    'assignment': 'tsc',
    'interlace': False,
    'catalogue_type': None,
    'statistic_type': None,
    'degrees': {'ell1': None, 'ell2': None, 'ELL': None},
    'wa_orders': {'i': None, 'j': None},
    'form': 'diag',
    'norm_convention': 'particle',
    'binning': 'lin',
    'range': [None, None],
    'num_bins': None,
    'idx_bin': None,
    'verbose': 20,
}


class NonParameterError(ValueError):
    """:exc:`ValueError` raised when a program parameter is non-existent.

    """
    pass


class InvalidParameterError(ValueError):
    """:exc:`ValueError` raised when a program parameter is invalid.

    """
    pass


cdef class ParameterSet:
    """Parameter set.

    This reads parameters from a file or a :class:`dict`, stores and
    prints out the extracted parameters, and validates the parameters.

    Parameters
    ----------
    param_filepath : str or :class:`pathlib.Path`, optional
        Parameter file path.  This should point to a YAML file.
        If `None` (default), `param_dict` should be provided; otherwise
        `param_dict` should be `None`.
    param_dict : dict, optional
        Parameter dictionary (nested).  If `None` (default),
        `param_filepath` should be provided; otherwise
        `param_filepath` should be `None`.
    logger : :class:`logging.Logger`, optional
        Program logger (default is `None`).

    Raises
    ------
    ValueError
        If neither or both of `param_filepath` and `param_dict`
        is/are `None`.


    .. tip::

        Use :func:`~triumvirate.parameters.fetch_paramset_template()` to
        generate a template parameter file (with explanatory text).

    """

    def __cinit__(self, param_filepath=None, param_dict=None, logger=None):

        self.thisptr = new CppParameterSet()

        self._logger = logger

        if (param_filepath is not None and param_dict is not None) \
                or (param_filepath is None and param_dict is None):
            raise ValueError(
                "One and only one of `param_filepath` and `param_dict` "
                "should be set."
            )

        self._params = {}
        if param_filepath is not None:
            self._source = str(Path(param_filepath).absolute())
            with open(param_filepath, 'r') as param_file:
                self._params = yaml.load(param_file, Loader=yaml.Loader)
        if param_dict is not None:
            self._source = 'dict'
            self._params = param_dict

        self._original = True
        self._validity = False

        self._parse_params()  # validate

    def __dealloc__(self):
        del self.thisptr

    def __str__(self):
        return "\n".join([
            f"ParameterSet(source={self._source}, original={self._original})",
            f"{pformat(self._params, sort_dicts=False)}",
        ])

    def __getitem__(self, key):
        """Return parameter value by key like :class:`dict`.

        Parameters
        ----------
        key : str
            Parameter name.

        Returns
        -------
        Parameter value.

        """
        try:
            return self._params[key]
        except KeyError:
            raise NonParameterError(f"Non-existent parameter: '{key}'.")

    def __setitem__(self, key, val):
        """Set parameter value by key like :class:`dict`.

        Parameters
        ----------
        key : str
            Parameter name.
        val :
            Parameter value.

        """
        self._params.update({key: val})
        self._parse_params()
        self._original = False

    def __getattr__(self, name):
        """Get parameter value as an attribute.

        Parameters
        ----------
        name : str
            Parameter name.

        Returns
        -------
        Parameter value.

        """
        try:
            return self.__getitem__(name)
        except (KeyError, NonParameterError) as err:
            raise AttributeError(str(err))

    def __setattr__(self, name, value):
        """Set parameter value as an attribute.

        Parameters
        ----------
        name : str
            Parameter name.
        value :
            Parameter value.

        """
        self.__setitem__(name, value)

    def names(self):
        """Return the set of top-level parameter names like
        :meth:`dict.keys`.

        Returns
        -------
        :obj:`dict_keys`
            Top-level parameter names as dictionary keys.

        """
        return self._params.keys()

    def items(self):
        """Return the set of entries like a :meth:`dict.items`.

        Returns
        -------
        :obj:`dict_items`
            Parameters as dictionary items.

        """
        return self._params.items()

    def get(self, key, *default_val):
        """Return a possibly non-existent top-level entry like
        :meth:`dict.get`.

        Parameters
        ----------
        key : str
            Top-level parameter name, possibly non-existent.
        default_val :
            Default value if the entry does not exist.

        Returns
        -------
        object
            Top-level parameter value, `None` if non-existent.

        """
        return self._params.get(key, *default_val)

    def update(self, *args, **kwargs):
        """Update parameter set like :meth:`dict.update`.

        """
        self._params.update(*args, **kwargs)
        self._parse_params()
        self._original = False

    def print(self):
        """Print the parameters as dictionary with :func:`pprint.pprint`.

        """
        pprint(self._params, sort_dicts=False)

    def save(self, filepath=None):
        """Save validated parameters to a YAML file.

        Parameters
        ----------
        filepath : str or :class:`pathlib.Path`, optional
            Saved file path.  If `None` (default), parameters are
            saved to a default file in the output measurement directory.

        """
        if filepath is None:
            filepath = Path(
                self._params['directories']['measurements'],
                "parameters_used{}.yml".format(self._params['tags']['output'])
            )

        with open(filepath, 'w') as printout_file:
            yaml.dump(
                self._params, printout_file,
                sort_keys=False,
                default_flow_style=False  # CAVEAT: discretionary
            )

        if self._logger:
            self._logger.info("Saved parameters to %s.", filepath)

    def _parse_params(self):
        """Parse the parameter set into the corresponding
        C++ parameter class.

        """
        # ----------------------------------------------------------------
        # Parsing
        # ----------------------------------------------------------------

        # -- I/O ---------------------------------------------------------

        # Directories, files and tags default to empty if unset.
        # Full path concatenation is handled by C++.

        # Attribute string parameters.
        try:
            if self._params['directories']['catalogues'] is None:
                self._params['directories']['catalogues'] = ""
                self.thisptr.catalogue_dir = "".encode('utf-8')
            else:
                self.thisptr.catalogue_dir = \
                    self._params['directories']['catalogues'].encode('utf-8')
        except KeyError:
            self.thisptr.catalogue_dir = "".encode('utf-8')

        try:
            if self._params['directories']['measurements'] is None:
                self._params['directories']['measurements'] = ""
                self.thisptr.measurement_dir = "".encode('utf-8')
            else:
                self.thisptr.measurement_dir = \
                    self._params['directories']['measurements'].encode('utf-8')
        except KeyError:
            self.thisptr.measurement_dir = "".encode('utf-8')

        try:
            if self._params['files']['data_catalogue'] is None:
                self._params['files']['data_catalogue'] = ""
                self.thisptr.data_catalogue_file = "".encode('utf-8')
            else:
                self.thisptr.data_catalogue_file = \
                    self._params['files']['data_catalogue'].encode('utf-8')
        except KeyError:
            self.thisptr.data_catalogue_file = "".encode('utf-8')

        try:
            if self._params['files']['rand_catalogue'] is None:
                self._params['files']['rand_catalogue'] = ""
                self.thisptr.rand_catalogue_file = "".encode('utf-8')
            else:
                self.thisptr.rand_catalogue_file = \
                    self._params['files']['rand_catalogue'].encode('utf-8')
        except KeyError:
            self.thisptr.rand_catalogue_file = "".encode('utf-8')

        try:
            if self._params['tags']['output'] is None:
                self._params['tags']['output'] = ''
                self.thisptr.output_tag = ''.encode('utf-8')
            else:
                self.thisptr.output_tag = \
                    self._params['tags']['output'].encode('utf-8')
        except KeyError:
            self.thisptr.output_tag = ''.encode('utf-8')

        # -- Mesh sampling -----------------------------------------------

        # Attribute numerical parameters.
        if None in self._params['boxsize'].values():
            raise InvalidParameterError("`boxsize` parameters must be set.")
        self.thisptr.boxsize = [
            float(self._params['boxsize']['x']),
            float(self._params['boxsize']['y']),
            float(self._params['boxsize']['z']),
        ]
        if None in self._params['ngrid'].values():
            raise InvalidParameterError("`ngrid` parameters must be set.")
        self.thisptr.ngrid = [
            self._params['ngrid']['x'],
            self._params['ngrid']['y'],
            self._params['ngrid']['z'],
        ]

        if self._params['padfactor'] is not None:
            self.thisptr.padfactor = self._params['padfactor']

        # Attribute string parameters.
        if self._params['alignment'] is not None:
            self.thisptr.alignment = \
                self._params['alignment'].lower().encode('utf-8')
        if self._params['padscale'] is not None:
            self.thisptr.padscale = \
                self._params['padscale'].lower().encode('utf-8')
        if self._params['assignment'] is not None:
            self.thisptr.assignment = \
                self._params['assignment'].lower().encode('utf-8')
        if self._params['interlace'] is not None:  # possibly convert from bool
            self.thisptr.interlace = \
                str(self._params['interlace']).lower().encode('utf-8')

        # Attribute derived parameters.
        self.thisptr.volume = np.prod(list(self._params['boxsize'].values()))
        self.thisptr.nmesh = np.prod(list(self._params['ngrid'].values()))

        # -- Measurement -------------------------------------------------

        # Attribute numerical parameters.
        if self._params['degrees']['ell1'] is not None:
            self.thisptr.ell1 = self._params['degrees']['ell1']
        if self._params['degrees']['ell2'] is not None:
            self.thisptr.ell2 = self._params['degrees']['ell2']
        if self._params['degrees']['ELL'] is not None:
            self.thisptr.ELL = self._params['degrees']['ELL']
        else:
            raise InvalidParameterError("`ELL` parameter must be set.")

        if self._params['wa_orders']['i'] is not None:
            self.thisptr.i_wa = self._params['wa_orders']['i']
        if self._params['wa_orders']['j'] is not None:
            self.thisptr.j_wa = self._params['wa_orders']['j']

        if self._params['range'] is None or None in self._params['range']:
            raise InvalidParameterError("`range` parameters must be set.")
        self.thisptr.bin_min = float(self._params['range'][0])
        self.thisptr.bin_max = float(self._params['range'][1])

        if self._params['num_bins'] is not None:
            self.thisptr.num_bins = self._params['num_bins']
        else:
            raise InvalidParameterError("`num_bins` parameter must be set.")
        if self._params['idx_bin'] is not None:
            self.thisptr.idx_bin = self._params['idx_bin']

        # Attribute string parameters.
        if self._params['catalogue_type'] is not None:
            self.thisptr.catalogue_type = \
                self._params['catalogue_type'].lower().encode('utf-8')
        # else:
        #     raise InvalidParameterError(
        #         "`catalogue_type` parameter must be set."
        #     )
        if self._params['statistic_type'] is not None:
            self.thisptr.statistic_type = \
                self._params['statistic_type'].lower().encode('utf-8')
        # else:
        #     raise InvalidParameterError(
        #         "`statistic_type` parameter must be set."
        #     )

        if self._params['form'] is not None:
            self.thisptr.form = \
                self._params['form'].lower().encode('utf-8')
        if self._params['form'] == 'full' and self._params['idx_bin'] is None:
            raise InvalidParameterError(
                "`idx_bin` parameter must be set when `form` is 'full'."
            )

        if self._params['norm_convention'] is not None:
            self.thisptr.norm_convention = \
                self._params['norm_convention'].lower().encode('utf-8')

        if self._params['binning'] is not None:
            self.thisptr.binning = \
                self._params['binning'].lower().encode('utf-8')

        # -- Misc --------------------------------------------------------

        if self._params['verbose'] is not None:
            self.thisptr.verbose = self._params['verbose']

        # ----------------------------------------------------------------
        # Validation
        # ----------------------------------------------------------------

        self._validate()

    def _validate(self):
        """Validate extracted parameters.

        """
        try:
            self._logger.info("Validating parameters...", cpp_state='start')
        except (AttributeError, TypeError):
            pass

        if self.thisptr.validate() != 0:
            raise InvalidParameterError("Invalid measurement parameters.")

        # Fetch validated parameters that have been derived or transmuted.
        # Transmuted I/O paths are not fetched.
        self._params['npoint'] = self.thisptr.npoint.decode('utf-8')
        self._params['space'] = self.thisptr.space.decode('utf-8')

        _interlace = self.thisptr.interlace.decode('utf-8')
        if _interlace.lower() == 'true':
            self._params['interlace'] = True
        elif _interlace.lower() == 'false':
            self._params['interlace'] = False
        else:
            raise RuntimeError(
                "C++ instance of trv::ParameterSet did not return "
                "a recognised 'interlace' string in {'true', 'false'}."
            )

        self._validity = True

        try:
            self._logger.info("... validated parameters.", cpp_state='end')
        except (AttributeError, TypeError):
            pass


def fetch_paramset_template(format, ret_defaults=False):
    """Fetch a parameter set template, either as a YAML file
    (with explanatory text) or as a dictionary.

    The returned template parameters are not validated as some mandatory
    values are unset.

    Parameters
    ----------
    format : {'text', 'dict'}
        Template format, either file contents ('text')
        or dictionary ('dict').
    ret_defaults : bool, optional
        If `True` (default is `False`), return a list of non-`None`
        default parameters in the template.  Only used when `format`
        is 'dict'.

    Returns
    -------
    template : str or dict
        Parameter set template as text or a dictionary.
    defaults : dict
        Parameter set default entries.  Only returned when `format` is
        'dict' and `ret_defaults` is `True`.

    Raises
    ------
    ValueError
        If `format` is neither 'text' nor 'dict'.

    """
    pkg_root_dir = Path(__file__).parent
    tml_filepath = pkg_root_dir/"resources"/"params_template.yml"

    text_template = tml_filepath.read_text()
    dict_template = yaml.load(text_template, Loader=yaml.Loader)

    if not ret_defaults:
        if format.lower() == 'text':
            return text_template
        if format.lower() == 'dict':
            return dict_template

    defaults = {}
    for key, val in dict_template.items():
        if isinstance(val, (tuple, list, set, dict)):
            continue
        if val is not None:
            defaults.update({key: val})

    if format.lower() == 'text':
        return text_template, defaults
    if format.lower() == 'dict':
        return dict_template, defaults

    raise ValueError("`format` must be either 'text' or 'dict'.")


def _modify_sampling_parameters(paramset, params_sampling=None,
                                params_default=None, ret_defaults=False):
    """Modify sampling parameters in a parameter set.

    Parameters
    ----------
    paramset : dict-like
        Parameter set to be updated.
    params_sampling : dict, optional
        Dictionary containing a subset of the following entries
        for sampling parameters---

            * 'boxsize': [float, float, float];
            * 'ngrid': [int, int, int];
            * 'alignment': {'centre', 'pad'}
            * 'assignment': {'ngp', 'cic', 'tsc', 'pcs'};
            * 'interlace': bool;

        and one and only one of the following when 'alignment' is 'pad'---

            * 'boxpad': float;
            * 'gridpad': float.

        This will override corresponding parameters in the template.
        If `None` (default), no modification happens and the original
        `paramset` is returned.
    params_default : dict, optional
        If not `None` (default), this is a sub-dictionary of `paramset`
        and any of its entries unmodified by `params_sampling` will be
        registered as default values.
    ret_defaults : bool, optional
        If `True` (default is `False`), return a list of non-`None`
        default parameters which have not been modified.  Only used if
        `params_default` is not `None`.

    Returns
    -------
    paramset : dict-like
        Modified parameter set.
    params_default : dict, optional
        Only returned when `ret_defaults` is `True` and there are
        unmodified default parameters from  `params_default`.

    Raises
    ------
    ValueError
        When `params_default` is not a sub-dictionary of `paramset`.

    """
    if params_sampling is None:
        return paramset

    if params_default is not None:
        if not (params_default.items() <= paramset.items()):
            raise ValueError(
                "`params_default` should be a subdictionary of `paramset`."
            )

    if 'alignment' in params_sampling.keys():
        paramset['alignment'] = params_sampling['alignment']
        if params_default: params_default.pop('alignment', None)

    if paramset.get('alignment') == 'pad':
        if 'boxpad' in params_sampling.keys() \
                and 'gridpad' in params_sampling.keys():
            raise ValueError(
                "Cannot set both 'boxpad` and 'gridpad' in `params_sampling`."
            )
    if 'boxpad' in params_sampling.keys():
        paramset['padscale'] = 'box'
        paramset['padfactor'] = params_sampling['boxpad']
        if params_default:
            params_default.pop('padscale', None)
            params_default.pop('padfactor', None)
    if 'gridpad' in params_sampling.keys():
        paramset['padscale'] = 'grid'
        paramset['padfactor'] = params_sampling['gridpad']
        if params_default:
            params_default.pop('padscale', None)
            params_default.pop('padfactor', None)

    if 'boxsize' in params_sampling.keys():
        paramset['boxsize'] = dict(zip(
            ['x', 'y', 'z'], params_sampling['boxsize']
        ))
        if params_default: params_default.pop('boxsize', None)
    if 'ngrid' in params_sampling.keys():
        paramset['ngrid'] = dict(zip(
            ['x', 'y', 'z'], params_sampling['ngrid']
        ))
        if params_default: params_default.pop('ngrid', None)
    if 'assignment' in params_sampling.keys():
        paramset['assignment'] = params_sampling['assignment']
        if params_default: params_default.pop('assignment', None)
    if 'interlace' in params_sampling.keys():
        paramset['interlace'] = params_sampling['interlace']
        if params_default: params_default.pop('interlace', None)

    if ret_defaults:
        return paramset, params_default
    return paramset


def _modify_measurement_parameters(paramset, params_measure=None,
                                   params_default=None, ret_defaults=False):
    """Modify measurement parameters in a parameter set.

    Parameters
    ----------
    paramset : dict-like
        Parameter dictionary to be updated.
    params_measure : dict, optional
        Dictionary containing a subset of the following entries
        for measurement parameters:

            * 'ell1', 'ell2', 'ELL': int;
            * 'i_wa', 'j_wa': int;
            * 'form': {'diag', 'full'};
            * 'bin_min', 'bin_max': float;
            * 'num_bins': int;
            * 'idx_bin': int.

        This will override corresponding parameters in the template.
        If `None` (default), no modification happens and the original
        `paramset` is returned.
    params_default : dict, optional
        If not `None` (default), this is a sub-dictionary of `paramset`
        and any of its entries unmodified by `params_measure` will be
        registered as default values.
    ret_defaults : bool, optional
        If `True` (default is `False`), return a list of non-`None`
        default parameters which have not been modified.  Only used if
        `params_default` is not `None`.

    Returns
    -------
    paramset : dict-like
        Modified parameter dictionary.
    params_default : dict, optional
        Only returned when `ret_defaults` is `True` and there are
        unmodified default parameters from `params_default`.

    Raises
    ------
    ValueError
        When `params_default` is not a subdictionary of `paramset`.

    """
    if params_measure is None:
        return paramset

    if params_default is not None:
        if not (params_default.items() <= paramset.items()):
            raise ValueError(
                "`params_default` should be a subdictionary of `paramset`."
            )

    for deg_label in ['ell1', 'ell2', 'ELL']:
        if deg_label in params_measure.keys():
            paramset['degrees'][deg_label] = params_measure[deg_label]

    for waorder_label in ['i_wa', 'j_wa']:
        if waorder_label in params_measure.keys():
            paramset['wa_orders'][waorder_label.rstrip('_wa')] = \
                params_measure[waorder_label]

    # Update 'idx_bin' before 'form' owing to dependency.
    if 'idx_bin' in params_measure.keys():
        paramset['idx_bin'] = params_measure['idx_bin']
        if params_default: params_default.pop('idx_bin', None)
    if 'form' in params_measure.keys():
        paramset['form'] = params_measure['form']
        if params_default: params_default.pop('form', None)

    if 'bin_min' in params_measure.keys():
        paramset['range'][0] = params_measure['bin_min']
    if 'bin_max' in params_measure.keys():
        paramset['range'][1] = params_measure['bin_max']
    if 'num_bins' in params_measure.keys():
        paramset['num_bins'] = params_measure['num_bins']
        if params_default: params_default.pop('num_bins', None)

    if ret_defaults:
        return paramset, params_default
    return paramset



"""
Double Hankel Transforms (:mod:`~triumvirate.bihankel`)
==========================================================================

Perform Hankel-related double transforms using the FFTLog algorithm.

Array operations
----------------

.. autosummary::
    ShapeError

1-d arrays:

.. autosummary::
    extrap_lin
    extrap_loglin
    extrap_pad

2-d arrays:

.. autosummary:
    extrap2d_bilin
    extrap2d_biloglin
    extrap2d_bipad

Transforms
----------

.. autosummary::
    transform_bispec_to_3pcf
    transform_3pcf_to_bispec

"""
import numpy as np
from scipy import interpolate

import triumvirate._fftlog as _fftlog


class ShapeError(ValueError):
    """Exception raised when the array shape is incorrect.

    """
    pass


# ========================================================================
# Extrapolation
# ========================================================================

def _check_1d_array(a):
    """Check the input array is a sorted 1-d array.

    Parameters
    ----------
    a : array of float
        Input 1-d array.

    Returns
    -------
    a : array of float
        Input 1-d array.

    Raises
    ------
    ShapeError
        When the input array is not 1-d.

    """
    a = np.squeeze(a)

    if a.ndim != 1:
        raise ShapeError("Input array is not 1-d.")
    # if not np.all(a[:-1] <= a[1:]):
    #     raise ValueError("Input array is not an increasing sequence.")

    return a


def _check_2d_array(a):
    """Check the input array is a sorted 2-d array.

    Parameters
    ----------
    a : array of float
        Input 2-d array.

    Returns
    -------
    a : array of float
        Input 2-d array.

    Raises
    ------
    ShapeError
        When the input array is not 2-d.

    """
    a = np.squeeze(a)
    if a.ndim != 2:
        raise ShapeError("Input array is not 1-d.")
    # if (
    #     not np.all([np.all(a_row[:-1] <= a_row[1:]) for a_row in a])
    # ) or (
    #     not np.all([
    #         np.all(a_col[:-1] <= a_col[1:]) for a_col in a.transpose()
    #     ])
    # ):
    #     raise ValueError("Input array is not an increasing sequence.")

    return a


def extrap_lin(a, n_ext):
    """Extrapolate a 1-d array linearly.

    The linear spacing used is the distance between the two points
    at either edge separately.

    Parameters
    ----------
    a : 1-d array of float
        Input 1-d array.
    n_ext : int
        Number of extra elements on either side.

    Returns
    -------
    a_out : :class:`numpy.ndarray`
        Extrapolated 1-d array.

    """
    a = _check_1d_array(a)

    a_l = a[0] + (a[1] - a[0]) * np.arange(-n_ext, 0)
    a_r = a[-1] + (a[-1] - a[-2]) * np.arange(1, n_ext + 1)
    a_out = np.concatenate([a_l, a, a_r])

    return a_out


def extrap_loglin(a, n_ext):
    """Extrapolate a 1-d array log-linearly.

    Parameters
    ----------
    a : 1-d array of float
        Inout 1-d array.
    n_ext : int
        Number of extra elements on either side.

    Returns
    -------
    a_out : :class:`numpy.ndarray`
        Extrapolated 1-d array.

    Raises
    ------
    ValueError
        When the input array contains non-positive entries.

    """
    a = _check_1d_array(a)
    if np.any(a <= 0.):
        raise ValueError("Input array contains non-positive entries.")

    a_l = a[0] * (a[1] / a[0]) ** np.arange(-n_ext, 0)
    a_r = a[-1] * (a[-1] / a[-2]) ** np.arange(1, n_ext + 1)
    a_out = np.concatenate([a_l, a, a_r])

    return a_out


def extrap_pad(a, n_ext, a_const=0.):
    """Extrapolate a 1-d array by constant padding.

    Parameters
    ----------
    a : 1-d array of float
        Input 1-d array.
    n_ext : int
        Number of extra elements on either side.
    a_const : float, optional
        Constant value (default is 0.) used for padding.

    Returns
    -------
    a_out : :class:`numpy.ndarray`
        Extrapolated 1-d array.

    """
    a = _check_1d_array(a)

    a_l = a_r = np.full(n_ext, a_const)
    a_out = np.concatenate([a_l, a, a_r])

    return a_out


def _extrap2d_lin(a, ncol_ext):
    """Extrapolate a 2-d array linearly along each row
    (i.e. horizontally).

    Parameters
    ----------
    a : 2-d :class:`numpy.ndarray`
        Input 2-d array.
    ncol_ext : int
        Number of extra columns on either side.

    Returns
    -------
    a_out : :class:`numpy.ndarray`
        Extrapolated 2-d array.

    """
    a_l = a[:, [0]] + np.arange(-ncol_ext, 0) * (a[:, [1]] - a[:, [0]])
    a_r = a[:, [-1]] + np.arange(1, ncol_ext + 1) * (a[:, [-1]] - a[:, [-2]])

    a_out = np.c_[a_l, a, a_r]

    return a_out


def _extrap2d_loglin(a, ncol_ext):
    """Extrapolate a 2-d array log-linearly along each row
    (i.e. horizontally).

    Parameters
    ----------
    a : 2-d :class:`numpy.ndarray`
        Input 2-d array.
    ncol_ext : int
        Number of extra columns on either side.

    Returns
    -------
    a_out : :class:`numpy.ndarray`
        Extrapolated 2-d array.

    """
    a_l = a[:, [0]] * (a[:, [1]] / a[:, [0]]) ** np.arange(-ncol_ext, 0)
    a_r = a[:, [-1]] * (a[:, [-1]] / a[:, [-2]]) ** np.arange(1, ncol_ext + 1)

    a_out = np.c_[a_l, a, a_r]

    return a_out


def extrap2d_bilin(a, ncol_ext, nrow_ext=None):
    """Extrapolate a 2-d array bilinearly.

    Parameters
    ----------
    a : 2-d array of float
        Input 2-d array.
    ncol_ext : int
        Number of extra columns on either side of each row.
    nrow_ext : int, optional
        Number of extra rows on either side of each column.  If `None`
        (default), this is set to `ncol_ext`.

    Returns
    -------
    a_out : :class:`numpy.ndarray`
        Extrapolated 2-d array.

    Raises
    ------
    ValueError
        If the input array is not equivalently 2-d.

    """
    a = _check_2d_array(a)

    # Extrapolate horizontally.
    a_ = _extrap2d_lin(a, ncol_ext)

    # Extrapolate vertically.  Use the same algorithm for the horizontal
    # extrapolation by transposing.
    nrow_ext = ncol_ext if nrow_ext is None else nrow_ext

    a_out = _extrap2d_lin(a_.transpose(), nrow_ext).transpose()

    return a_out


def extrap2d_logbilin(a, ncol_ext, nrow_ext=None):
    """Extrapolate a 2-d array log-bilinearly.

    Parameters
    ----------
    a : 2-d array of float
        Input 2-d array.
    ncol_ext : int
        Number of extra columns on either side of each row.
    nrow_ext : int, optional
        Number of extra rows on either side of each column.  If `None`
        (default), this is set to `ncol_ext`.

    Returns
    -------
    a_out : :class:`numpy.ndarray`
        Extrapolated 2-d array.

    Raises
    ------
    ValueError
        When the input array contains non-positive entries.

    """
    a = _check_2d_array(a)
    if np.any(a <= 0.):
        raise ValueError("Input array contains non-positive entries.")

    # Extrapolate horizontally.
    a_ = _extrap2d_loglin(a, ncol_ext)

    # Extrapolate vertically.  Use the same algorithm for the horizontal
    # extrapolation by transposing.
    nrow_ext = ncol_ext if nrow_ext is None else nrow_ext

    a_out = _extrap2d_loglin(a_.transpose(), nrow_ext).transpose()

    return a_out


def extrap2d_bipad(a, ncol_ext, nrow_ext=None,
                   a_const_col=0., a_const_row=None):
    """Extrapolate a 2-d array by constant padding.

    Parameters
    ----------
    a : 2-d array of float
        Input 2-d array.
    ncol_ext : int
        Number of extra columns on either side of each row.
    nrow_ext : int, optional
        Number of extra rows on either side of each column.  If `None`
        (default), this is set to `ncol_ext`.
    a_const_col : float, optional
        Constant value (default is 0.) used for padding in each row.
    a_const_row : float, optional
        Constant value used for padding in each column.  If `None`
        (default), this is set to `a_const_col`.

    Returns
    -------
    a_out : :class:`numpy.ndarray`
        Extrapolated 2-d array.

    Raises
    ------
    ValueError
        If the input array is not equivalently 2-d.

    """
    a = _check_2d_array(a)

    nrow_ext = ncol_ext if nrow_ext is None else nrow_ext
    a_const_row = a_const_col if a_const_row is None else a_const_row

    a_l = a_r = np.full((a.shape[0], ncol_ext), a_const_col)
    a_ = np.c_[a_l, a, a_r]

    a__u = a__d = np.full((nrow_ext, a_.shape[-1]), a_const_row)
    a_out = np.r_[a__u, a_, a__d]

    return a_out


# ========================================================================
# Transforms
# ========================================================================

def transform_bispec_to_3pcf(ell1, ell2, k_in, bk_in, r_out,
                             n_fftlog, extrap=None, n_extrap=None):
    r"""Transform bispectrum samples to three-point correlation function
    (3PCF) samples using double 1-d FFTLog operations.

    Parameters
    ----------
    ell1, ell2 : int
        Multipole degree of the bispectrum.
    k_in : 1-d array of float
        Input wavenumbers.
    bk_in : 2-d array of float
        Input bispectrum.
    r_out : 1-d array of float
        Output separations.
    n_fftlog : int
        FFTLog sample number.
    extrap : {'lin', 'loglin', 'zero'}, optional
        If not `None` (default), set one of the following options---

        - 'lin': input bispectrum is extrapolated linearly;
        - 'loglin': input bispectrum is extrapolated log-linearly;
        - 'zero': input bispectrum is zero padded.

        In each case, `n_extrap` must be set, and the input wavenumbers
        `k_in` are extrapolated log-linearly.
    n_extrap : int, optional
        Extrapolation sample number (one-sided) (default is `None`).

    Returns
    -------
    zeta_dict : dict
        Output 3PCF including its attributes---

        - 'r', 'zeta': post-transform separation and 3PCF samples;
        - 'r_fftlog', 'zeta_fftlog': FFTLog separation and 3PCF samples;
        - 'r1_mesh', 'r2_mesh': post-transform separation mesh grid;
        - 'n_fftlog': FFTLog sample number;
        - 'ell1', 'ell2': multipole degrees.

    Raises
    ------
    ValueError
        When `n_extrap` is not set but `extrap` is.


    .. attention::

        This function assumes the wavenumber samples :math:`k_{1,2}` in
        the bispectrum :math:`B_{\ell_1 \ell_2 L}(k_1, k_2)` are the same.

    """
    # Prepare samples.
    if extrap is not None:
        if n_extrap is None:
            raise ValueError("`n_extrap` must be provided if `extrap` is set.")
        k_sample = extrap_loglin(k_in, n_extrap)

        if extrap == 'lin':
            bk_sample = extrap2d_bilin(bk_in, n_extrap)
        elif extrap == 'loglin':
            bk_sample = extrap2d_logbilin(bk_in, n_extrap)
        elif extrap == 'zero':
            bk_sample = extrap2d_bipad(bk_in, n_extrap)
        else:
            raise ValueError(f"Unknown `extrap` option: {extrap}.")
    else:
        k_sample, bk_sample = k_in, bk_in

    n_sample = len(k_sample)

    # Perform FFTLog transform.
    # CAVEAT: The choice of interpolator is discretionary but should
    # have no effect.
    k_fftlog = np.logspace(*np.log(k_sample[[0, -1]]), n_fftlog, base=np.e)

    zeta_mesh_partial = np.zeros((n_sample, n_fftlog))
    for i in range(n_sample):
        interpolator_bk = interpolate.interp1d(
            k_sample, bk_sample[i, :],
            fill_value='extrapolate', kind='cubic'
        )
        bk_fftlog = interpolator_bk(k_fftlog)

        r_fftlog, zeta_fftlog = _fftlog.sj_transform(
            ell2, 2, n_fftlog, k_fftlog, bk_fftlog
        )

        zeta_mesh_partial[i, :] = zeta_fftlog

    zeta_mesh_full = np.zeros((n_fftlog, n_fftlog))
    for j in range(n_fftlog):
        interpolator_bk = interpolate.interp1d(
            k_sample, zeta_mesh_partial[:, j],
            fill_value='extrapolate', kind='cubic'
        )
        bk_fftlog = interpolator_bk(k_fftlog)

        r_fftlog, zeta_fftlog = _fftlog.sj_transform(
            ell1, 2, n_fftlog, k_fftlog, bk_fftlog
        )

        zeta_mesh_full[:, j] = zeta_fftlog

    # An overall complex `parity` is needed here as it is not included
    # in the backend transforms to avoid redundant computation.
    parity = np.real(1j ** (ell1 + ell2))
    zeta_fftlog = parity * zeta_mesh_full

    # Prepare output.
    interpolator2d_zeta = interpolate.RectBivariateSpline(
        r_fftlog, r_fftlog, zeta_fftlog
    )

    zeta_out = interpolator2d_zeta(r_out, r_out)

    (r2_mesh, r1_mesh) = np.meshgrid(r_out, r_out)

    zeta_dict = {
        'r': r_out,
        'zeta': zeta_out,
        'r_fftlog': r_fftlog,
        'zeta_fftlog': zeta_fftlog,
        'r1_mesh': r1_mesh,
        'r2_mesh': r2_mesh,
        'n_fftlog': n_fftlog,
        'ell1': ell1,
        'ell2': ell2,
    }

    return zeta_dict


def transform_3pcf_to_bispec(ell1, ell2, r_in, zeta_in, k_out,
                             n_fftlog, extrap=None, n_extrap=None):
    r"""Transform three-point correlation function (3PCF) samples to
    bispectrum samples using double 1-d FFTLog operations.

    Parameters
    ----------
    ell1, ell2 : int
        Multipole degree of the 3PCF.
    r_in : 1-d array of float
        Input separations.
    zeta_in : 2-d array of float
        Input 3PCF.
    k_out : 1-d array of float
        Output wavenumbers.
    n_fftlog : int
        FFTLog sample number.
    extrap : {'lin', 'loglin', 'zero'}, optional
        If not `None` (default), set one of the following options:

        - 'lin': input 3PCF is extrapolated linearly;
        - 'loglin': input 3PCF is extrapolated log-linearly;
        - 'zero': input 3PCF is zero padded.

        In each case, `n_extrap` must be set, and the input separations
        `r_in` are extrapolated log-linearly.
    n_extrap : int, optional
        Extrapolation sample number (one-sided) (default is `None`).

    Returns
    -------
    bk_dict : dict
        Output bispectrum including its attributes:

        - 'k', 'bk': post-transform wavenumber and bispectrum samples;
        - 'k_fftlog', 'bk_fftlog': FFTLog wavenumber and
          bispectrum samples;
        - 'k1_mesh', 'k2_mesh': post-transform wavenumber mesh grid;
        - 'n_fftlog': FFTLog sample number;
        - 'ell1', 'ell2': multipole degrees.

    Raises
    ------
    ValueError
        When `n_extrap` is not set but `extrap` is.


    .. attention::

        This function assumes the separation samples :math:`r_{1,2}` in
        the 3PCF :math:`\zeta_{\ell_1 \ell_2 L}(r_1, r_2)` are the same.

    """
    # Prepare samples.
    if extrap is not None:
        if n_extrap is None:
            raise ValueError("`n_extrap` must be provided if `extrap` is set.")
        r_sample = extrap_loglin(r_in, n_extrap)

        if extrap == 'lin':
            zeta_sample = extrap2d_bilin(zeta_in, n_extrap)
        elif extrap == 'loglin':
            zeta_sample = extrap2d_logbilin(zeta_in, n_extrap)
        elif extrap == 'zero':
            zeta_sample = extrap2d_bipad(zeta_in, n_extrap)
        else:
            raise ValueError(f"Unknown `extrap` option: {extrap}.")
    else:
        r_sample, zeta_sample = r_in, zeta_in

    n_sample = len(r_sample)

    # Perform FFTLog transform.
    # CAVEAT: The choice of interpolator is discretionary but should
    # have no effect.
    r_fftlog = np.logspace(*np.log(r_sample[[0, -1]]), n_fftlog, base=np.e)

    bk_mesh_partial = np.zeros((n_sample, n_fftlog))
    for i in range(n_sample):
        interpolator_zeta = interpolate.interp1d(
            r_sample, zeta_sample[i, :],
            fill_value='extrapolate', kind='cubic'
        )
        zeta_fftlog = interpolator_zeta(r_fftlog)

        k_fftlog, bk_fftlog = _fftlog.sj_transform(
            ell2, 2, n_fftlog, r_fftlog, zeta_fftlog
        )

        bk_mesh_partial[i, :] = bk_fftlog

    bk_mesh_full = np.zeros((n_fftlog, n_fftlog))
    for j in range(n_fftlog):
        interpolator_zeta = interpolate.interp1d(
            r_sample, bk_mesh_partial[:, j],
            fill_value='extrapolate', kind='cubic'
        )
        zeta_fftlog = interpolator_zeta(r_fftlog)

        k_fftlog, bk_fftlog = _fftlog.sj_transform(
            ell1, 2, n_fftlog, r_fftlog, zeta_fftlog
        )

        bk_mesh_full[:, j] = bk_fftlog

    # An overall complex `parity` is needed here as it is not included
    # in the backend transforms to avoid redundant computation.
    # Factors of pi are needed here because the forward transform is used
    # in the backend as the backward transform.
    parity = np.real((-1j) ** (ell1 + ell2))
    bk_fftlog = (2*np.pi) ** 6 * parity * bk_mesh_full

    # Prepare output.
    interpolator2d_bk = interpolate.RectBivariateSpline(
        k_fftlog, k_fftlog, bk_fftlog
    )

    bk_out = interpolator2d_bk(k_out, k_out)

    (k2_mesh, k1_mesh) = np.meshgrid(k_out, k_out)

    bk_dict = {
        'k': k_out,
        'bk': bk_out,
        'k_fftlog': k_fftlog,
        'bk_fftlog': bk_fftlog,
        'k1_mesh': k1_mesh,
        'k2_mesh': k2_mesh,
        'n_fftlog': n_fftlog,
        'ell1': ell1,
        'ell2': ell2,
    }

    return bk_dict



"""Test :mod:`~triumvirate.twopt`.

"""
import numpy as np
import pytest

from triumvirate.twopt import (
    compute_corrfunc,
    compute_corrfunc_in_gpp_box,
    compute_corrfunc_window,
    compute_powspec,
    compute_powspec_in_gpp_box
)


@pytest.mark.slow
@pytest.mark.parametrize(
    "degree",
    [0, 2,]  # noqa: E231
)
def test_compute_powspec(degree,
                         test_data_catalogue, test_rand_catalogue,
                         test_binning_fourier,
                         test_paramset,
                         test_logger,
                         test_stats_dir):

    measurements = compute_powspec(
        test_data_catalogue, test_rand_catalogue,
        degree=degree,
        binning=test_binning_fourier,
        paramset=test_paramset,
        logger=test_logger
    )
    measurements_ext = np.loadtxt(
        test_stats_dir/f"pk{degree}_lpp.txt", unpack=True
    )

    assert np.allclose(measurements['kbin'], measurements_ext[0]), \
        "Measurement bins do not match."
    assert np.allclose(measurements['keff'], measurements_ext[1]), \
        "Measured coordinates do not match."
    assert np.allclose(measurements['nmodes'], measurements_ext[2]), \
        "Measured mode counts do not match."
    assert np.allclose(
        measurements['pk_raw'],
        measurements_ext[3] + 1j * measurements_ext[4]
    ), "Measured raw statistics do not match."
    assert np.allclose(
        measurements['pk_shot'],
        measurements_ext[5] + 1j * measurements_ext[6],
        atol=1.e-6
    ), "Measured shot noise contributions do not match."


@pytest.mark.slow
@pytest.mark.parametrize(
    "degree",
    [0, 2,]  # noqa: E231
)
def test_compute_corrfunc(degree,
                          test_data_catalogue, test_rand_catalogue,
                          test_binning_config,
                          test_paramset,
                          test_logger,
                          test_stats_dir):

    measurements = compute_corrfunc(
        test_data_catalogue, test_rand_catalogue,
        degree=degree,
        binning=test_binning_config,
        paramset=test_paramset,
        logger=test_logger
    )
    measurements_ext = np.loadtxt(
        test_stats_dir/f"xi{degree}_lpp.txt", unpack=True
    )

    assert np.allclose(
        measurements['rbin'], measurements_ext[0]
    ), "Measurement bins do not match."
    assert np.allclose(
        measurements['reff'], measurements_ext[1]
    ), "Measured coordinates do not match."
    assert np.allclose(
        measurements['npairs'], measurements_ext[2]
    ), "Measured pair counts do not match."
    assert np.allclose(
        measurements['xi'],
        measurements_ext[3] + 1j * measurements_ext[4]
    ), "Measured statistics do not match."


@pytest.mark.slow
@pytest.mark.parametrize(
    "degree",
    [0, 2,]  # noqa: E231
)
def test_compute_powspec_in_gpp_box(degree,
                                    test_data_catalogue,
                                    test_binning_fourier,
                                    test_paramset,
                                    test_logger,
                                    test_stats_dir):

    measurements = compute_powspec_in_gpp_box(
        test_data_catalogue,
        degree=degree,
        binning=test_binning_fourier,
        paramset=test_paramset,
        logger=test_logger
    )
    measurements_ext = np.loadtxt(
        test_stats_dir/f"pk{degree}_gpp.txt", unpack=True
    )

    assert np.allclose(measurements['kbin'], measurements_ext[0]), \
        "Measurement bins do not match."
    assert np.allclose(measurements['keff'], measurements_ext[1]), \
        "Measured coordinates do not match."
    assert np.allclose(measurements['nmodes'], measurements_ext[2]), \
        "Measured mode counts do not match."
    assert np.allclose(
        measurements['pk_raw'],
        measurements_ext[3] + 1j * measurements_ext[4]
    ), "Measured raw statistics do not match."
    assert np.allclose(
        measurements['pk_shot'],
        measurements_ext[5] + 1j * measurements_ext[6],
        atol=1.e-6
    ), "Measured shot noise contributions do not match."


@pytest.mark.slow
@pytest.mark.parametrize(
    "degree",
    [0, 2,]  # noqa: E231
)
def test_compute_corrfunc_in_gpp_box(degree,
                                     test_data_catalogue,
                                     test_binning_config,
                                     test_paramset,
                                     test_logger,
                                     test_stats_dir):

    measurements = compute_corrfunc_in_gpp_box(
        test_data_catalogue,
        degree=degree,
        binning=test_binning_config,
        paramset=test_paramset,
        logger=test_logger
    )
    measurements_ext = np.loadtxt(
        test_stats_dir/f"xi{degree}_gpp.txt", unpack=True
    )

    assert np.allclose(
        measurements['rbin'], measurements_ext[0]
    ), "Measurement bins do not match."
    assert np.allclose(
        measurements['reff'], measurements_ext[1]
    ), "Measured coordinates do not match."
    assert np.allclose(
        measurements['npairs'], measurements_ext[2]
    ), "Measured pair counts do not match."
    assert np.allclose(
        measurements['xi'],
        measurements_ext[3] + 1j * measurements_ext[4]
    ), "Measured statistics do not match."


@pytest.mark.slow
@pytest.mark.parametrize(
    "degree",
    [0, 2,]  # noqa: E231
)
def test_compute_corrfunc_window(degree,
                                 test_rand_catalogue,
                                 test_binning_config,
                                 test_paramset,
                                 test_logger,
                                 test_stats_dir):

    measurements = compute_corrfunc_window(
        test_rand_catalogue,
        degree=degree,
        binning=test_binning_config,
        paramset=test_paramset,
        logger=test_logger
    )
    measurements_ext = np.loadtxt(
        test_stats_dir/f"xiw{degree}.txt", unpack=True
    )

    assert np.allclose(
        measurements['rbin'], measurements_ext[0]
    ), "Measurement bins do not match."
    assert np.allclose(
        measurements['reff'], measurements_ext[1]
    ), "Measured coordinates do not match."
    assert np.allclose(
        measurements['npairs'], measurements_ext[2]
    ), "Measured pair counts do not match."
    assert np.allclose(
        measurements['xi'],
        measurements_ext[3] + 1j * measurements_ext[4]
    ), "Measured statistics do not match."



"""
Catalogue Parser (:mod:`~triumvirate._particles`)
==========================================================================

Parse Python catalogue objects into C++ particle catalogues.

"""
cimport numpy as np

from ._particles cimport CppParticleCatalogue


cdef class _ParticleCatalogue:

    def __cinit__(
        self,
        np.ndarray[double, ndim=1, mode='c'] x not None,
        np.ndarray[double, ndim=1, mode='c'] y not None,
        np.ndarray[double, ndim=1, mode='c'] z not None,
        np.ndarray[double, ndim=1, mode='c'] nz not None,
        np.ndarray[double, ndim=1, mode='c'] ws not None,
        np.ndarray[double, ndim=1, mode='c'] wc not None,
        verbose=-1
    ):

        self.thisptr = new CppParticleCatalogue(verbose)

        self.thisptr.load_particle_data(x, y, z, nz, ws, wc)

    def __dealloc__(self):
        del self.thisptr



"""Set up Triumvirate and its Cythonised extension modules.

"""
import logging
import os
import platform
import sys
from copy import deepcopy
from multiprocessing import cpu_count
from setuptools import setup
from setuptools.command.build_clib import build_clib

import numpy
from Cython.Build import cythonize
from Cython.Distutils import build_ext, Extension
from extension_helpers._openmp_helpers import check_openmp_support
from extension_helpers._setup_helpers import pkg_config


distutils_logger = logging.getLogger()  # recycled from :mod:`distutils._log`


# ========================================================================
# Package
# ========================================================================

PROJECT_NAME = 'Triumvirate'


def get_pkg_name():
    """Get package name in lower-case format.

    Returns
    -------
    str
        Package name.

    """
    return PROJECT_NAME.lower()


def get_pkg_dir(topdir="src"):
    """Get package directory.

    Parameters
    ----------
    topdir : str, optional
        Top directory.

    Returns
    -------
    str
        Package directory.

    """
    return os.path.join(topdir, get_pkg_name())


def get_pkg_include_dir(subdir="include"):
    """Get package C++ header directory.

    Parameters
    ----------
    subdir : str, optional
        Subdirectory of C++ headers in the package directory.

    Returns
    -------
    str
        C++ header directory.

    """
    return os.path.join(get_pkg_dir(), subdir)


def get_pkg_src_dir(subdir="src/modules"):
    """Get package C++ source directory.

    Parameters
    ----------
    subdir : str, optional
        Subdirectory of C++ sources in the package directory.

    Returns
    -------
    str
        C++ source directory.

    """
    return os.path.join(get_pkg_dir(), subdir)


# ========================================================================
# Build
# ========================================================================

class BuildExt(build_ext):
    """Modified :class:`Cython.Distutils.build_ext`.

    """
    def finalize_options(self):
        """Modify parallelisation option in
        :meth:`Cython.Distutils.build_ext.build_ext.build_extensions`.

        """
        super().finalize_options()

        _num_procs = get_build_num_procs()
        if self.parallel is None and _num_procs is not None:
            self.parallel = _num_procs

    def build_extensions(self):
        """Modify compiler and compilation configuration in
        :meth:`Cython.Distutils.build_ext.build_ext.build_extensions`.

        """
        OPTS_TO_REMOVE = ['-Wstrict-prototypes',]  # noqa: E231
        for opt in OPTS_TO_REMOVE:
            try:
                self.compiler.compiler_so.remove(opt)
            except ValueError:
                pass

        try:
            _num_procs = max(int(self.parallel), 1)
        except TypeError:
            _num_procs = 1

        distutils_logger.info(
            "running build_ext on %d process(es) with %s compiler",
            _num_procs, self.compiler.compiler_so[0],
        )

        super().build_extensions()


class BuildClib(build_clib):
    """Modified :class:`setuptools.command.build_clib.build_clib`.

    """
    def build_libraries(self, libraries):
        """Modify compiler and compilation configuration in
        :meth:`setuptools.command.build_clib.build_clib.build_libraries`.

        """
        OPTS_TO_REMOVE = ['-Wstrict-prototypes',]  # noqa: E231
        for opt in OPTS_TO_REMOVE:
            try:
                self.compiler.compiler_so.remove(opt)
            except ValueError:
                pass

        distutils_logger.info(
            "running build_clib with %s compiler",
            self.compiler.compiler_so[0],
        )

        super().build_libraries(libraries)


def get_build_num_procs():
    """Get build parallel processes.

    Returns
    -------
    int or None
        Number of parallel processes, which may be `None` in the case
        of no parallelisation.

    """
    flag_parallel = os.environ.get('PY_BUILD_PARALLEL', '').strip()
    if '-j' == flag_parallel:
        num_procs = cpu_count()
    elif '-j' in flag_parallel:
        try:
            num_procs = int(flag_parallel.lstrip('-j'))
        except ValueError:
            num_procs = None
    else:
        num_procs = None

    return num_procs


def prioprint(*args, **kwargs):
    """`print` in high priority.

    """
    print(*args, file=sys.stderr, **kwargs)


def display_py_environs():
    """Display customised environment variables passed to setup.

    """
    PY_ENV_VARS = [
        'PY_CXX',
        'PY_CXXFLAGS',
        'PY_LDFLAGS',
        'PY_INCLUDES',
        'PY_NO_OMP',
        'PY_CXXFLAGS_OMP',
        'PY_LDFLAGS_OMP',
        'PY_BUILD_PARALLEL',
    ]
    for env_var in PY_ENV_VARS:
        prioprint("{}={}".format(env_var, os.environ.get(env_var)))


def display_py_options():
    """Display customised compilation options used in build.

    """
    PY_OPT_TYPES = [
        'macros',
        'cflags',
        'ldflags',
        'libs',
        'include_dirs',
        'lib_dirs',
    ]
    for opt_type in PY_OPT_TYPES:
        prioprint("{}={}".format(opt_type, globals().get(opt_type)))


# ========================================================================
# Compilation
# ========================================================================

EXT_LANG = 'c++'


# -- Environment ---------------------------------------------------------

def get_platform():
    """Get the build platform.

    Returns
    -------
    {'linux', 'darwin', 'default'}

    """
    build_platform = platform.system().lower()
    if build_platform not in ['linux', 'darwin']:
        build_platform = 'default'

    return build_platform


# -- Compiler ------------------------------------------------------------

# List default compilers (GCC assumed).
COMPILERS = {
    'default': 'g++',
    'linux': 'g++',
    'darwin': 'g++',
}


def get_compiler():
    """Get the build compiler.

    Returns
    -------
    str
        Build compiler.

    """
    compiler = os.environ.get('PY_CXX', COMPILERS[get_platform()])

    return compiler


def set_cli_compiler(compiler=None):
    """Set command-line compiler through the ``CC``and ``CXX``
    environmental variables.

    """
    _compiler = compiler or get_compiler()

    os.environ['CC'] = _compiler
    os.environ['CXX'] = _compiler


# -- Dependencies --------------------------------------------------------

# Default to GCC OpenMP implementation.
OPENMP_LIBS = {
    'default': 'gomp',  # 'omp'
    'linux': '',  # set by ``-fopenmp`` or environment
    'darwin': '',  # set by ``-fopenmp`` or environment
}


LIBS_CORE = ['gsl', 'fftw3',]  # noqa: E231
LIBS_FULL = ['gsl', 'gslcblas', 'm', 'fftw3',]  # noqa: E231

PKG_LIB_NAME = 'trv'


# -- Options -------------------------------------------------------------

def convert_macro(macro):
    """Convert a macro flag to a tuple.

    Parameters
    ----------
    macro : str
        Macro as a flag

    Returns
    -------
    tuple (str, Union[str, None])
        Macro as a tuple

    """
    macro_ = macro.lstrip('-D').split('=')
    if len(macro_) == 1:
        return (macro_.pop(), None)
    if len(macro_) == 2:
        return tuple(macro_)
    raise ValueError(
        "Invalid macro to convert: {}.".format(macro_)
    )


def parse_cli_cflags():
    """Parse command-line ``PY_CXXFLAGS`` components for extension
    keyword arguments.

    Returns
    -------
    list of str, list of str
        Parsed `macros` and `cflags`.

    """
    cli_cflags = os.environ.get('PY_CXXFLAGS', '').split()

    parsed_macros, parsed_cflags = [], []
    for cflag_ in cli_cflags:
        if cflag_.startswith('-D'):
            parsed_macros.append(convert_macro(cflag_))
        else:
            parsed_cflags.append(cflag_)

    return parsed_macros, parsed_cflags


def parse_cli_ldflags():
    """Parse command-line ``PY_LDFLAGS`` components for extension
    keyword arguments.

    Returns
    -------
    list of str, list of str, list of str
        Parsed `ldflags`, `libs` and `libdirs`.

    """
    cli_ldflags = os.environ.get('PY_LDFLAGS', '').split()

    parsed_ldflags, parsed_libs, parsed_lib_dirs = [], [], []
    for ldflag_ in cli_ldflags:
        if ldflag_.startswith('-l'):
            parsed_libs.append(ldflag_.lstrip('-l'))
        elif ldflag_.startswith('-L'):
            parsed_lib_dirs.append(ldflag_.lstrip('-L'))
        else:
            parsed_ldflags.append(ldflag_)

    return parsed_ldflags, parsed_libs, parsed_lib_dirs


def parse_cli_includes():
    """Parse command-line ``PY_INCLUDES`` components for extension
    keyword arguments.

    Returns
    -------
    list of str
        Parsed `ldflags`.

    """
    parsed_include_dirs = \
        os.environ.get('PY_INCLUDES', '').replace('-I', '').split()

    return parsed_include_dirs


def parse_cli_cflags_omp():
    """Parse command-line ``PY_CXXFLAGS_OMP`` components for extension
    keyword arguments.

    """
    cli_cflags_omp = os.environ.get('PY_CXXFLAGS_OMP', '').split()
    if not cli_cflags_omp:
        return None

    parsed_macros_omp, parsed_cflags_omp = [], []
    for cflag_ in cli_cflags_omp:
        if cflag_.startswith('-D'):
            macro_ = cflag_.lstrip('-D').split('=')
            if len(macro_) == 1:
                parsed_macros_omp.append((macro_.pop(), None))
            elif len(macro_) == 2:
                parsed_macros_omp.append(tuple(macro_))
            else:
                raise ValueError(
                    "Invalid macro in CXXFLAG_OMP: {}.".format(cflag_)
                )
        else:
            parsed_cflags_omp.append(cflag_)

    return parsed_macros_omp, parsed_cflags_omp


def parse_cli_ldflags_omp():
    """Parse command-line ``PY_LDFLAGS_OMP`` components for extension
    keyword arguments.

    """
    cli_ldflags_omp = os.environ.get('PY_LDFLAGS_OMP', '').split()
    if not cli_ldflags_omp:
        return None

    parsed_ldflags_omp, parsed_libs_omp, parsed_lib_dirs_omp = [], [], []
    for ldflag_ in cli_ldflags_omp:
        if ldflag_.startswith('-l'):
            parsed_libs_omp.append(ldflag_.lstrip('-l'))
        elif ldflag_.startswith('-L'):
            parsed_lib_dirs_omp.append(ldflag_.lstrip('-L'))
        else:
            parsed_ldflags_omp.append(ldflag_)

    return parsed_ldflags_omp, parsed_libs_omp, parsed_lib_dirs_omp


def add_options_nonomp(macros, cflags, ldflags, libs, lib_dirs, include_dirs):
    """Add any missing non-OpenMP compilation options.

    Parameters
    ----------
    macros : list of tuple (str, Union[str, None])
        Macros without '-D' prefix.
    cflags : list of str
        ``CXXFLAGS`` components.
    ldflags : list of str
        ``LDFLAGS`` components with non-'-l' prefixes.
    libs : list of str
        Libraries without the '-l' prefix.
    lib_dirs : list of str
        Library directories without the '-L' prefix.
    include_dirs : list of str
        ``INCLUDES`` directories without the '-I' prefix.

    Returns
    -------
    macros : list of tuple (str, Union[str, None])
        Macros without '-D' prefix.
    cflags : list of str
        ``CXXFLAGS`` components.
    ldflags : list of str
        ``LDFLAGS`` components with non-'-l' prefixes.
    libs : list of str
        Libraries without the '-l' prefix.
    lib_dirs : list of str
        Library directories without the '-L' prefix.
    include_dirs :list of str
        ``INCLUDES`` directories without the '-I' prefix.

    """
    libs_core = deepcopy(LIBS_CORE)
    libs_full = deepcopy(LIBS_FULL)

    for lib_ in libs:
        if lib_ not in libs_full:
            libs_core.append(lib_)
            libs_full.append(lib_)

    checked_options = pkg_config(libs_core, default_libraries=libs_full)

    for macro_ in checked_options['define_macros']:
        if macro_ not in macros:
            macros.append(macro_)
    for macro_ in checked_options['undef_macros']:
        if macro_ not in macros:
            macros.append(macro_)
    for flag in checked_options['extra_compile_args']:
        if flag not in cflags + ldflags:
            cflags.append(flag)
    for include_dir_ in checked_options['include_dirs']:
        if include_dir_ not in include_dirs:
            include_dirs.append(include_dir_)
    for lib_dir_ in checked_options['library_dirs']:
        if lib_dir_ not in lib_dirs:
            lib_dirs.append(lib_dir_)

    libs = checked_options['libraries']

    return macros, cflags, ldflags, libs, lib_dirs, include_dirs


def add_options_omp(macros, cflags, ldflags, libs, lib_dirs, include_dirs):
    """Add OpenMP options, if enabled.

    By default, GCC OpenMP is assumed.

    Parameters
    ----------
    macros : list of tuple (str, Union[str, None])
        Macros without '-D' prefix.
    cflags : list of str
        ``CXXFLAGS`` components.
    ldflags : list of str
        ``LDFLAGS`` components with non-'-l' prefixes.
    libs : list of str
        Libraries without the '-l' prefix.
    lib_dirs : list of str
        Library directories without the '-L' prefix.
    include_dirs :list of str
        ``INCLUDES`` directories without the '-I' prefix.

    Returns
    -------
    macros : list of tuple (str, Union[str, None])
        Processed macros without '-D' prefix.
    cflags : list of str
        Processed ``CXXFLAGS`` components.
    ldflags : list of str
        Processed ``LDFLAGS`` components non-'-l' prefix.
    libs : list of str
        Libraries without the '-l' prefix.
    lib_dirs : list of str
        Library directories without the '-L' prefix.
    include_dirs :list of str
        ``INCLUDES`` directories without the '-I' prefix.

    """
    # Check if OpenMP is explicitly disabled.
    if os.environ.get('PY_NO_OMP') is not None:
        prioprint("OpenMP is disabled explicitly.")
        return macros, cflags, ldflags, libs, lib_dirs, include_dirs

    # Check if OpenMP is unsupported.
    if not check_openmp_support():
        prioprint(
            "OpenMP is disabled as "
            "it is not supported in this build enviromnent."
        )
        return macros, cflags, ldflags, libs, lib_dirs, include_dirs

    # Enable OpenMP by default otherwise.
    prioprint("OpenMP is enabled by default.")

    # Adapt macros and `cflags`.
    MACROS_OMP_RELATED = [
        ('TRV_USE_OMP', None),
        ('TRV_USE_FFTWOMP', None),
    ]
    for macro_ in MACROS_OMP_RELATED:
        if macro_ not in macros:
            macros.append(macro_)

    try:
        macros_omp, cflags_omp = parse_cli_cflags_omp()
    except TypeError:
        macros_omp = []
        cflags_omp = ['-fopenmp',]  # noqa: E231

    for macro_ in macros_omp:
        macro_ = convert_macro(macro_)
        if macro_ not in macros:
            macros.append(macro_)
    for cflag_ in cflags_omp:
        if cflag_ not in cflags:
            cflags.append(cflag_)

    # Adapt `ldflags`, `libs` and `lib_dirs`.
    LIBS_OMP_BASED = [
        'fftw3_omp',
    ]  # noqa: E231
    for lib_ in LIBS_OMP_BASED:
        if lib_ and lib_ not in libs:
            libs.append(lib_)

    try:
        ldflags_omp, libs_omp, lib_dirs_omp = parse_cli_ldflags_omp()
    except TypeError:
        ldflags_omp = ['-fopenmp',]  # noqa: E231
        libs_omp = [OPENMP_LIBS[get_platform()],]  # noqa: E231
        lib_dirs_omp = []  # noqa: E231

    for ldflag_ in ldflags_omp:
        if ldflag_ not in ldflags:
            ldflags.append(ldflag_)
    for lib_ in libs_omp:
        if lib_ and lib_ not in libs:
            libs.append(lib_)
    for lib_dir_ in lib_dirs_omp:
        if lib_dir_ not in lib_dirs:
            lib_dirs.append(lib_dir_)

    return macros, cflags, ldflags, libs, lib_dirs, include_dirs


def add_options_pkgs(macros, cflags, ldflags, libs, lib_dirs, include_dirs):
    """Add options required by this package and external packages.

    Parameters
    ----------
    macros : list of tuple (str, Union[str, None])
        Macros without '-D' prefix.
    cflags : list of str
        ``CXXFLAGS`` components.
    ldflags : list of str
        ``LDFLAGS`` components with non-'-l' prefixes.
    libs : list of str
        Libraries without the '-l' prefix.
    lib_dirs : list of str
        Library directories without the '-L' prefix.
    include_dirs : list of str
        ``INCLUDES`` directories without the '-I' prefix.

    Returns
    -------
    macros : list of tuple (str, Union[str, None])
        Processed macros without '-D' prefix.
    cflags : list of str
        Processed ``CXXFLAGS`` components.
    ldflags : list of str
        Processed ``LDFLAGS`` components non-'-l' prefix.
    libs : list of str
        Libraries without the '-l' prefix.
    lib_dirs : list of str
        Library directories without the '-L' prefix.
    include_dirs : list of str
        ``INCLUDES`` directories without the '-I' prefix.

    """
    # Adapt `macros`.
    MACROS_PKG = [
        ('TRV_EXTCALL', None),
        # ('TRV_USE_LEGACY_CODE', None),
        # ('DBG_MODE', None),
        # ('DBG_FLAG_NOAC', None),
    ]
    MACROS_EXT = [
        ('NPY_NO_DEPRECATED_API', 'NPY_1_7_API_VERSION'),
    ]

    for macro_ in MACROS_PKG + MACROS_EXT:
        macros.append(macro_)

    # Adapt `libs`. Note the linking order matters.
    libs = [PKG_LIB_NAME,] + libs  # noqa: E231

    # Adapt `include_dirs`.
    include_dirs_pkg = [
        os.path.abspath(get_pkg_include_dir()),
    ]
    include_dirs_ext = [
        numpy.get_include(),
    ]

    for include_dir_ in include_dirs_pkg + include_dirs_ext:
        if include_dir_ not in include_dirs:
            include_dirs.append(include_dir_)

    return macros, cflags, ldflags, libs, lib_dirs, include_dirs


# ========================================================================
# Targets
# ========================================================================

CYTHON_DIRECTIVES = {
   'language_level': '3',
   'c_string_encoding': 'utf-8',
   'embedsignature': True,
}

EXT_CONFIGS = {
    'parameters': {},
    'dataobjs': {},
    '_particles': {},
    '_twopt': {},
    '_threept': {},
    '_fftlog': {},
}


def define_pkg_library(lib_name=PKG_LIB_NAME):
    """Define package library to be built and linked against.

    Parameters
    ----------
    lib_name : str
        Package libaray name.

    Returns
    -------
    tuple (str, dict)
        Package library and corresponding configurations.

    """
    pkg_src_dir = get_pkg_src_dir()

    pkg_sources = [
        os.path.join(pkg_src_dir, _cpp_source)
        for _cpp_source in os.listdir(pkg_src_dir)
    ]

    pkg_cfg = {
        cfg_opt: globals()[cfg_opt]
        for cfg_opt in ['macros', 'cflags', 'include_dirs',]  # noqa: E231
    }

    pkg_library = (lib_name, dict(sources=pkg_sources, **pkg_cfg))

    return pkg_library


def define_pkg_extension(ext_name,
                         auto_cpp_source=False, extra_cpp_sources=None,
                         **ext_kwargs):
    """Define package extension from given source files and options.

    Parameters
    ----------
    ext_name : str
        Extension module name in the form ``<pkg_name>.<ext_name>``.
        This sets the .pyx source file to ``<pkg_dir>/<ext_name>.pyx``.
    auto_cpp_source : bool, optional
        If `True` (default is `False`), add
        ``<pkg_src_dir>/<ext_name>.cpp`` to the list of source files
        with any prefix underscores removed from ``<ext_name>``.
    extra_cpp_sources : list of str, optional
        Additional .cpp source files under ``<pkg_src_dir>``
        (default is `None`).
    **ext_kwargs
        Options to pass to :class:`Cython.Distutils.Extension`.

    Returns
    -------
    :class:`Cython.Distutils.Extension`
        Cython extension.

    """
    EXT_CFG_MAPPING = [
        ('define_macros', 'macros'),
        ('extra_compile_args', 'cflags'),
        ('extra_link_args', 'ldflags'),
        ('include_dirs', 'include_dirs'),
        ('library_dirs', 'lib_dirs'),
        ('libraries', 'libs'),
    ]

    # Define Cython sources.
    pkg_dir = get_pkg_dir()

    sources = [os.path.join(pkg_dir, f"{ext_name}.pyx"),]  # noqa: E231

    # Add C++ sources.
    pkg_src_dir = get_pkg_src_dir()

    if auto_cpp_source:
        sources.append(
            os.path.join(pkg_src_dir, f"{ext_name}.cpp".lstrip('_'))
        )

    if extra_cpp_sources is None:
        extra_cpp_sources = []
    for _cpp_source in extra_cpp_sources:
        sources.append(os.path.join(pkg_src_dir, _cpp_source))

    # Determine extension configurations.
    ext_module_kwargs = {
        cfg: globals()[cfg_var]
        for (cfg, cfg_var) in EXT_CFG_MAPPING
    }

    if ext_kwargs:
        ext_module_kwargs.update(ext_kwargs)

    return Extension(
        '{}.{}'.format(get_pkg_name(), ext_name),
        sources=sources, language=EXT_LANG, **ext_module_kwargs
    )


# ========================================================================
# Set-up
# ========================================================================

if __name__ == '__main__':

    # Check build environment.
    display_py_environs()

    # Parse compilation options.
    set_cli_compiler()

    macros, cflags = parse_cli_cflags()
    ldflags, libs, lib_dirs = parse_cli_ldflags()
    include_dirs = parse_cli_includes()

    # Adapt compilation options.
    macros, cflags, ldflags, libs, lib_dirs, include_dirs = add_options_nonomp(
        macros, cflags, ldflags, libs, lib_dirs, include_dirs
    )
    macros, cflags, ldflags, libs, lib_dirs, include_dirs = add_options_omp(
        macros, cflags, ldflags, libs, lib_dirs, include_dirs
    )
    macros, cflags, ldflags, libs, lib_dirs, include_dirs = add_options_pkgs(
        macros, cflags, ldflags, libs, lib_dirs, include_dirs
    )

    # Check compilation options.
    display_py_options()

    # Define build targets.
    pkg_libraries = [define_pkg_library(),]  # noqa: E231

    pkg_extensions = [
        define_pkg_extension(ext, **cfg)
        for ext, cfg in EXT_CONFIGS.items()
    ]

    # Run build commands.
    cython_ext_modules = cythonize(
        pkg_extensions,
        compiler_directives=CYTHON_DIRECTIVES,
        nthreads=get_build_num_procs(),
    )

    setup(
        use_scm_version={
            'version_scheme': 'post-release',
            # 'local_scheme': 'no-local-version',
        },
        cmdclass={
            'build_clib': BuildClib,
            'build_ext': BuildExt,
        },
        ext_modules=cython_ext_modules,
        libraries=pkg_libraries,
    )
